!     Last change:  WRG  27 Apr 2005    2:46 am
MODULE AMOEBAMOD
!
! Adapted from Numerical Recipes by Martin Burda
! Last modified: 06/08/04
!
! Notes: 1. Specific interface for the objective function with its real name 
!           needs to be provided either in the calling program or in an INTERFACE 
!           module USEd by the calling program.
!        2. That same interface (but with the name "func" has to be copied 
!           into this subroutine (2 times), in lines 48-54, and 138-144. 
!        3. The only actual arguments of the objective function should be
!           the free parameters xx over which the search is performed. All other
!           data entering the function should be USE associated from another data 
!           module. 
       
 IMPLICIT NONE
   INTERFACE swap
      MODULE PROCEDURE swap_s,swap_v
   END INTERFACE
 
 CONTAINS

 SUBROUTINE AMOEBA(func,xx,xint,fmin,ftol,iter,grid,pxint)
!================================================================================!
!                         Subroutine AMOEBA                                      !
!================================================================================!
! Finds global minium of a user-supplied N-dim function using a grid of vectors  !
! as starting points for each individual simplex searches.                       !
! The grid can be optionally switched off to perform only 1 simplex search.      !
!                                                                                !
! INPUT : func : user-defined external function to be minimized                  ! 
!         xx   : N-dim vector, the initial argmin guess                          !
!         xint : N-dim vector, interval to be searched in each dim               !
!         ftol : convergence tolerance level (e.g. 10.d-12)                      !
!         grid : if ==.true. then xint split around xx into a grid in each dim   ! 
!                if ==.false. then xint taken only around xx (1 search)          !
!         pxint: (optional) = fraction of xint taken in one grid step            ! 
!                default pxint(:) = 0.1d0                                        !
! OUTPUT: fmin : value of the minimized function                                 !
!         xx   : (returns the argmin vector)                                     ! 
!         iter : number of amoeba_nr iterations for the best start grid vector   !
! CALLS : SINGLE_AM (subroutine for 1 simplex search)                            ! 
!--------------------------------------------------------------------------------!
  REAL(8), INTENT(INOUT)       :: xx(:)
  REAL(8), INTENT(IN)          :: xint(:),ftol
  LOGICAL, INTENT(IN)          :: grid
  REAL(8), INTENT(OUT)         :: fmin
  INTEGER, INTENT(OUT)         :: iter
  REAL(8), INTENT(IN),OPTIONAL :: pxint(:)
  INTERFACE
     FUNCTION func(x) 
    !USE GLOBAL_DATA
     IMPLICIT NONE
     REAL(8),INTENT(IN) :: x(:)
     REAL(8)            :: func 
     END FUNCTION func
  END INTERFACE 
  REAL(8)                      :: p(size(xx)+1,size(xx)),fminbest
  REAL(8), DIMENSION(size(xx)) :: xini,stsz,pxintd,xintd,pminbest,xxd
  INTEGER, DIMENSION(size(xx)) :: snt,giter
  INTEGER                      :: ii,jj,dim,Tgiter,iterbest,nt

   dim = size(xx)
   IF (PRESENT(pxint)) THEN                         ! dummy for pxint of size(xx)
      pxintd = pxint                               
   ELSE
      pxintd = 0.1d0                                ! set default if pxint not present
   END IF
   IF (grid) THEN                                   ! START grid amoeba
      giter  = int(1.d0/pxint)                      ! # of iter in each dimension
      stsz   = xint*pxint                           ! step size in each dimension
      xx     = xx-0.5d0*xint                        ! starting position in a grid corner
      xini   = xx                                   ! record initial value
      Tgiter = 1                                    ! initialize total # of grid steps             
      snt    = 0                                    ! initialize step counter
      DO ii=1,dim                                   ! determine total # of grid steps
         Tgiter = Tgiter*giter(ii)
      END DO
      xintd    = xint*pxint                         ! shrink xint for each step in grid
      fminbest = HUGE(1.d0)                         ! initialize
      DO nt=1,Tgiter                                ! go through the grid       
         CALL CREATE_SIMPLEX
         CALL SINGLE_AM(func,p,fmin,ftol,iter)
         IF (fmin<fminbest) THEN                    ! record if new min found
            fminbest = fmin       
            pminbest = p(1,:)
            iterbest = iter
         END IF
         xx(1)  = xx(1)+stsz(1)                     ! move up dim 1
         snt(1) = snt(1)+1                          ! steps taken in dim 1
         DO jj=1,dim                                ! check all dimensions for boundary
            IF (snt(jj)>=giter(jj)) THEN            ! if boundary hit
                xx(jj)  = xini(jj)                  ! return finished dim to ini value
                snt(jj) = 0                         ! reset the counter to steps+1       
                IF (jj<dim) THEN                    ! if not the  last dimension
                   xx(jj+1)  = xx(jj+1)+stsz(jj+1)  ! move next dim one step up 
                   snt(jj+1) = snt(jj+1) + 1        ! stepcount in next dimension
                END IF
            END IF            
         END DO
      END DO
      fmin = fminbest
      xx   = pminbest
      iter = iterbest                               ! FINISH grid amoba

   ELSE                                             ! START a single-simplex search 
!      WRITE(*,*)''
!      WRITE(*,*)'starting simplex search around a single vector'
!      WRITE(*,*)''
      xintd = xint
      CALL CREATE_SIMPLEX
      CALL SINGLE_AM(func,p,fmin,ftol,iter)
      xx = p(1,:)
   END IF
 CONTAINS 
      !-------------------------
       SUBROUTINE CREATE_SIMPLEX                    ! rows of p are the simplex verteces       
         INTEGER :: ii2
           xxd = xx - 0.5d0*xintd                   !  go back half interval       
           DO ii2=1,dim+1                           !  initialize all verteces (rows)
              p(ii2,:) = xxd              
           END DO
           DO ii2=1,dim                             !  strech one dim of each vertex at a time  
              p(ii2+1,ii2) = xxd(ii2) + 1.d0*xintd(ii2) 
           END DO
       END SUBROUTINE CREATE_SIMPLEX
      !-------------------------
END SUBROUTINE AMOEBA


SUBROUTINE SINGLE_AM(func,p,fmin,ftol,iter)
!---------------------------------------------------
! Subroutine SINGLE_AM
! Minimization of a real-valued function in N-dim
! by the downhill simplex method of Nelder and Mead. 
!---------------------------------------------------

	INTEGER, INTENT(OUT)   :: iter
	REAL(8), INTENT(IN)    :: ftol
	REAL(8), INTENT(OUT)   :: fmin
	REAL(8), INTENT(INOUT) :: p(:,:)
        INTERFACE
           FUNCTION func(x) 
          !USE GLOBAL_DATA
           IMPLICIT NONE
           REAL(8),INTENT(IN) :: x(:)
           REAL(8)            :: func 
           END FUNCTION func
        END INTERFACE 
	INTEGER, PARAMETER     :: ITMAX = 1000000
	REAL(8), PARAMETER     :: TINY  = 1.0d-10
	INTEGER                :: ihi,ndim,ct
	REAL(8)                :: psum(size(p,2)),y(size(p,2)+1),yprev
        
        ndim = assert_eq2(size(p,2),size(p,1)-1,'amoeba') 
        ! setup the initial funcional values of siplex verteces
          DO ct=1,ndim+1 
             y(ct) = func(p(ct,:))
          END DO
	call SUBSINGLE_AM

  CONTAINS
  !---------------------------------------
    SUBROUTINE SUBSINGLE_AM

	INTEGER :: i,ilo,inhi
	REAL(8) :: rtol,ysave,ytry,ytmp

            iter    = 0
            psum(:) = sum(p(:,:),dim=1)
            yprev = TINY
            DO
		ilo    = iminloc(y(:))
		ihi    = imaxloc(y(:))
		ytmp   = y(ihi)
		y(ihi) = y(ilo)
		inhi   = imaxloc(y(:))     
		y(ihi) = ytmp              
		rtol   = 2.d0*abs(y(ihi)-y(ilo))/(abs(y(ihi))+abs(y(ilo))+TINY)
                yprev  = y(ilo) 
		IF (rtol < ftol) then
			call swap(y(1),y(ilo))
                        call swap(p(1,:),p(ilo,:))
                        fmin = y(1)
                        RETURN
		END IF
		IF (iter >= ITMAX) THEN 
                   WRITE(*,*)'ITMAX exceeded in amoeba, min not achieved'
                   RETURN
                END IF
		ytry = inflex(-1.d0)
		iter = iter+1
		IF (ytry <= y(ilo)) THEN
			ytry = inflex(2.d0)
			iter = iter+1
		ELSE IF (ytry >= y(inhi)) THEN
			ysave = y(ihi)
			ytry  = inflex(0.5d0)
			iter  = iter+1
			IF (ytry >= ysave) THEN
				p(:,:) = 0.5d0*(p(:,:)+spread(p(ilo,:),1,size(p,1)))
				DO i=1,ndim+1
					IF (i /= ilo) y(i)=func(p(i,:))
				END DO
				iter    = iter+ndim
				psum(:) = sum(p(:,:),dim=1)
			END IF
		END IF
             END DO
   END SUBROUTINE SUBSINGLE_AM
     !--------------------------------------
     FUNCTION INFLEX(fac)
	REAL(8), INTENT(IN) :: fac
	REAL(8)             :: inflex
	REAL(8)             :: fac1,fac2,ytry
	REAL(8)             :: ptry(size(p,2))
	    fac1 = (1.d0-fac)/ndim 
            fac2 = fac1-fac        
            ptry(:) = psum(:)*fac1-p(ihi,:)*fac2 
            ytry = func(ptry)
            IF (ytry < y(ihi)) THEN
		y(ihi)  = ytry
		psum(:) = psum(:)-p(ihi,:)+ptry(:)
		p(ihi,:)= ptry(:)
            END IF
	    inflex = ytry
      END FUNCTION INFLEX
      !---------------------------------------
  END SUBROUTINE SINGLE_AM

!======================================================================
! Utility functions for amoeba 
  !---------------------------------------
	FUNCTION iminloc(arr)
	REAL(8), INTENT(IN) :: arr(:)
	INTEGER             :: imin(1)
	INTEGER             :: iminloc
	   imin=minloc(arr(:))
	   iminloc=imin(1)
	END FUNCTION iminloc
  !----------------------------------------
	FUNCTION imaxloc(arr)
	REAL(8), INTENT(IN) :: arr(:)
	INTEGER             :: imaxloc
	INTEGER             :: imax(1)
	   imax=maxloc(arr(:))
	   imaxloc=imax(1)
	END FUNCTION imaxloc
  !----------------------------------------
	SUBROUTINE swap_v(a,b)
	REAL(8), INTENT(INOUT) :: a(:),b(:)
	REAL(8)                :: dum(size(a))
	   dum = a
	   a   = b
	   b   = dum
	END SUBROUTINE swap_v 
  !----------------------------------------
	SUBROUTINE swap_s(a,b)
	REAL(8), INTENT(INOUT) :: a,b
	REAL(8)                :: dum 
	   dum = a
	   a   = b
	   b   = dum
	END SUBROUTINE swap_s
  !----------------------------------------
	FUNCTION assert_eq2(n1,n2,string)
	CHARACTER(LEN=*), INTENT(IN) :: string
	INTEGER,          INTENT(IN) :: n1,n2
	INTEGER                      :: assert_eq2
	   if (n1 == n2) then
		   assert_eq2=n1
           else
		  write (*,*) 'error: assert_eq failed with this tag:', &
			       string
                  write (*,*) 'inappropriate dimensions of arguments passed'
		  STOP 'program terminated by assert_eq3'
	   end if
	END FUNCTION assert_eq2
  !-----------------------------------------
END MODULE AMOEBAMOD
