!     Last change:  WRG   2 Jun 2054    6:59 pm
MODULE quadrature
 IMPLICIT NONE
 REAL(8),PARAMETER:: PI=3.14159265359d0

  CONTAINS

    SUBROUTINE gauleg(x1,x2,qa,qw)

    !****************************************************************************
    ! Given the lower and upper limits of integration x1 and x2 this routine    *
    ! returns arrays qa and qw of length N containing the abscissas and weights *
    ! of the Gauss-Legendre N-point quadrature formula.                         *
    !                                                                           *
    ! Adopted from the Numerical recipes for fortran 90                         *
    ! Wayne-Roy Gayle Oct 2001                                                  *
    !****************************************************************************

     IMPLICIT NONE
     REAL(8),INTENT(IN)::            x1,x2
     REAL(8),INTENT(OUT)::           qa(:),qw(:)
     REAL(8), PARAMETER ::           eps=3d-14
     INTEGER(4), PARAMETER ::        maxit=10
     INTEGER(4) ::                   its,j,m,n,i
     REAL(8) ::                      xl,xm,w
     REAL(4) ::                      y1(size(qa)), y2(size(qa))
     REAL(8), DIMENSION((size(qa)+1)/2) :: p1,p2,p3,pp,z,z1,temp
     LOGICAL, DIMENSION((size(qa)+1)/2) :: unfinished

     n=size(qa)
     m=(n+1)/2                                       ! The roots are symmetric in the interval,
     xm=0.5d0*(x2+x1)                                ! so we only have to .nd half of them. 
     xl=0.5d0*(x2-x1)
     call barth(temp,1.d0,1.d0,m)
     z=cos(PI*(temp-0.25d0)/(n+0.5d0))        ! Initial approximations to the roots.
     unfinished=.true.

     do its=1,MAXIT 
        where (unfinished)
            p1=1.0
            p2=0.0
        end where
        do j=1,n
            where (unfinished)
                p3=p2
                p2=p1
                p1=((2.0d0*j-1.0d0)*z*p2-(j-1.0d0)*p3)/j
            end where
        end do
        where (unfinished)
            pp=n*(z*p1-p2)/(z*z-1.0d0)
            z1=z
            z=z1-p1/pp
            unfinished=(abs(z-z1) > EPS)
        end where
            if (.not. any(unfinished)) exit
     end do
     if (its == MAXIT+1) then
        print *, 'too many iterations'
     end if
     qa(1:m)=xm-xl*z
     qa(n:n-m+1:-1)=xm+xl*z
     qw(1:m)=2.0d0*xl/((1.0d0-z**2)*pp**2)
     qw(n:n-m+1:-1)=qw(1:m)

     RETURN
    END SUBROUTINE gauleg



    SUBROUTINE  barth (arth,first,increment,n)
    !********************************************************************************
    ! Returns an array of length n containing an arithmetic progression whose       *
    ! first value is first and whose increment is increment. If first and           *
    ! increment have rank greater than zero, returns an array of one larger rank,   *
    ! with the last subscript having size n and indexing the progressions. Note that*
    ! the following reference implementation (for the scalar case) is definitional  *
    ! only, and neither parallelized nor optimized for roundoff error.              *
    !********************************************************************************

     IMPLICIT NONE
     REAL(8), INTENT(IN) ::        first,increment
     INTEGER(4), INTENT(IN) ::     n
     REAL(8), INTENT(OUT) ::       arth(n)
     INTEGER(4) ::                 k

     if (n > 0) arth(1)=first
     do k=2,n
       arth(k)=arth(k-1)+increment
     end do
     RETURN
    END SUBROUTINE barth


END MODULE quadrature
